import tkinter as tk
from tkinter import ttk
from tkinter import filedialog as fd
import natsort as nt
from datetime import datetime
from .configuration import config


class resultsDialog:
    """A dialog box to show completed calibration results per beam."""

    def __init__(self, parent, data: dict=None, icon=None):
        
        self.data = data
        
        self.top = tk.Toplevel(parent)
        self.top.title("Results")
        if icon:
            self.top.iconphoto(False, icon)

        tree_frame = ttk.Frame(self.top)

        # use a ttk.Treeview to show a table of the results
        self.item_ids = {}  # contains ids to rows that get added to the treeview
        self.setup_treeview(tree_frame, self.data)  # creates self.tree

        # Make scrollbars for the treeview widget
        vsb = ttk.Scrollbar(tree_frame, orient="vertical", command=self.tree.yview)
        self.tree.configure(yscrollcommand=vsb.set)
        
        # pack the scrollbars and treeview
        vsb.pack(side="right", fill="y")
        self.tree.pack(side=tk.TOP, fill=tk.BOTH, expand=True, padx=5)

        # create and pack the buttons
        btn_frame = ttk.Frame(self.top)
        remove = ttk.Button(btn_frame, text="Remove selected", command=self.remove_rows)
        save = ttk.Button(btn_frame, text="Save", command=self.save)
        close = ttk.Button(btn_frame, text="Close", command=self.close_dialog)
        
        close.pack(side=tk.RIGHT)
        save.pack(side=tk.RIGHT)
        remove.pack(side=tk.RIGHT)

        tree_frame.pack(side=tk.TOP, fill=tk.BOTH, expand=True)
        btn_frame.pack(side=tk.BOTTOM, fill=tk.X)

    def setup_treeview(self, top, data):
        """Create the treeview columns, etc."""
        
        headings = [data.df().index.name] + list(data.df().columns)
        
        self.tree = ttk.Treeview(top, columns=headings, show='headings')
        self.tree.bind('<Button-1>', self.on_row_click)

        # colour odd and even rows
        self.tree.tag_configure('evenrow', background='white smoke')
        self.tree.tag_configure('oddrow', background='white')
        self.tree.tag_configure('active', background=config.calibrating_colour())
        
        for col in headings:
            self.tree.heading(col, text=col, anchor='e')
            self.tree.column(col, width=125, anchor='e')

        # Add rows (if any)
        self.update_with(data)            
                    
    def update_with(self, data, active_beam_label: str = ''):
        """Update dialog's display with given calibration data."""
        
        def format(value):
            """Format floats to 1 decimal place."""
            if isinstance(value, float):
                return f'{value:0.1f}'
            return f'{value}'

        for beam_label, row in data.df().iterrows():
            values = [beam_label] + [format(v) for v in row]

            if beam_label in self.item_ids:
                # row for this beam already exists, so update it
                self.tree.item(self.item_ids[beam_label], values=values)
            else:
                # new beam, so add a row
                item_id = self.tree.insert('', 'end', values=values)
                self.item_ids[beam_label] = item_id

        # sort rows by beam_label and highlight the active beam (if one is active)
        self.update_rows(active_beam_label)

        # keep this to use in the save and remove functionalities
        self.data = data

    def update_rows(self, active_beam_label: str):
        """Sorts calibration results rows by beam and sets background colours to look nice."""

        for beam_label, index in zip(nt.natsorted(self.item_ids.keys()), range(len(self.item_ids))):
            self.tree.move(self.item_ids[beam_label], '', index)

            # Decide on the row colour
            rowness = 'oddrow' if index % 2 == 0 else 'evenrow'
            if active_beam_label and beam_label == active_beam_label:
                rowness = 'active'  # gets highlighted

            # Set the row colour in the tree widget
            self.tree.item(self.item_ids[beam_label], tags=(rowness,))    

    def remove_rows(self):
        """Remove selected rows from the results table."""

        selected = self.tree.selection()  # returns a tuple of item_ids
        
        # Get the beam labels of the selected rows
        to_remove = []
        for iid in selected:
            item_data = self.tree.item(iid)
            values = item_data['values']
            # Convert to str because the self.items_id map has str has the key (this
            # comes from the DataFrame this class has been given).
            to_remove.append(str(values[0]))

        if to_remove:
            for beam_label in to_remove:
                # remove selected rows from treeview
                self.tree.delete(self.item_ids[beam_label])
                # remove beam imtem from the map between beam and item ids
                self.item_ids.pop(beam_label)
            # remove beams from the cal_data store
            self.data.remove(to_remove)
            self.update_rows(None)

    def save(self):
        """Save the results to a file."""
        timestamp = datetime.now().strftime('%Y%m%dT%H%M%S')
        default_filename = 'sonar_calibration_' + timestamp + '.csv'
        save_filename = fd.asksaveasfilename(title='Save as CSV', defaultextension='.csv',
                                             initialdir=config.userDocumentsDir(),
                                             initialfile=default_filename,
                                             filetypes=[('CSV', '*.csv')])
        if save_filename:
            self.data.save(save_filename)

    def on_row_click(self, event):
        """Implement selection and deselection."""
        # TODO - work out why it can take multiple clicks on a row to get it unselected
        item_id = self.tree.identify_row(event.y)
        if item_id and item_id in self.tree.selection():
            self.tree.selection_remove(item_id)


    def reopen(self):
        self.top.deiconify()

    def close_dialog(self):
        self.top.withdraw()



"""Code to process Simrad raw sonar datagrams into Sv and TS."""
import logging
import numpy as np
from .utils import cartesian_to_spherical
from .configuration import config as config

logger = logging.getLogger(config.appName())


class rawDatagramProcessor():
    """Calculates Sv and TS from Simrad raw datagrams from sonars."""

    def __init__(self):
        self._raw_dgs = []
        self._power = []

        # There are multiple fans. 
        # In the example data:
        #
        # SN90 has 'Horizontal-H', 'Vertical-H', and 'InspectionC-H'
        # CS90 has 'Horizontal-H', and 'Vertical-H'
        #
        # This code currently only uses the horizontal fan
        self.selected_fan_name = 'Horizontal-H'
        self.selected_ping_name = 'Horizontal'

        self.product_name = ''
        self.sv = None  # [dB]
        self.ts = None  # [dB]
        self.sound_speed = None  # [m/s]
        self.ping_time = None  # [datetime]
        self.sample_interval = None  # [m]
        self.pulse_duration = None  # [s]
        self.frequency = None  # [Hz]
        self.transmit_power = None  # [W]
        self.equivalent_beam_angle = None  # [dB]
        self.labels = None
        self.theta = None  # [rad]
        self.tilts = None  # [rad]
        self.gain_rx = None  # [dB]
        self.gain_adjust = None  # [dB]
        self.sa_correction = None  # [dB]
        self.sa_correction_adjust = None  # [dB]
        self.absorption_coefficient = None  # [dB/m]
        
    def add_datagram(self, dg: dict) -> bool:
        """Accumulates datagrams for a ping.

        Parameters
        ----------
        dg :
            A Simrad sonar datagram
        
        Returns
        -------
        : True if all pings for a datagram have been received and processed ping data
            are available, otherwise False
        """
        
        if dg['type'] == 'EOP0':
            # have now received all data for a ping
            self._calculate_sv_ts()

            # in prep for new data
            self._power.clear()

            return True
        else:
            # Pick out data that doesn't need end of ping processing
            match dg['type']:
                case 'VER0':
                    self.product_name = dg['product_name']
                case 'PCO0' | 'PCO1':  # ping configuration, once per file
                    self._extract_ping_config(dg)
                case 'PIN0' | 'PIN1':  # ping information, once per ping
                    self.sound_speed = dg['sound_velocity']
                    self.ping_time = dg['ping_time'].datetime
                case 'RAW2':  # multiple per ping
                    self._accumulate_raw(dg)

            return False

    def _calculate_sv_ts(self) -> None:
        """Calculate Sv and TS for the just finished ping."""
        
        # reshape the power data to join together the separate blocks of power data
        power = np.concatenate(self._power, axis=1)
        num_samples = power.shape[1]

        gain_tx = 0  # unknown value so default to zero

        wavenumber = self.sound_speed / self.frequency
        tilt_corr = 30.0 *np.log10(np.cos(self.tilts))
        range_corr = 3.0  # [m] empirical range correction
        sample_range = self.sound_speed * self.sample_interval / 2.0
        r = np.arange(num_samples) * sample_range - range_corr
        # Ranges less than 0 get set to the smallest positive range
        r[r<sample_range] = sample_range

        tvg = 20*np.log10(r) + 2*r*self.absorption_coefficient
        sv_const = (10*np.log10(self.transmit_power * wavenumber**2 * self.sound_speed
                                * self.pulse_duration / (32.0 * np.pi**2)))\
                    + gain_tx + self.gain_rx + tilt_corr\
                    + 2*self.sa_correction\
                    + self.equivalent_beam_angle

        # make tvg and sv_const into 2D matrices to get broadcasting to work
        self.sv = power + tvg[np.newaxis, :] - sv_const[:, np.newaxis]

        tvg = 40*np.log10(r) + 2*r*self.absorption_coefficient
        ts_const = (10*np.log10(self.transmit_power * wavenumber**2 / (16.0 * np.pi**2)))\
                    + gain_tx + self.gain_rx + tilt_corr

        self.ts = power + tvg[np.newaxis, :] - ts_const[:, np.newaxis]

    def _accumulate_raw(self, raw: dict) -> None:
        
        # the raw datagrams can be from different beams and are also split into 
        # separate blocks (split by sample it seems). It also seems they they always
        # arrive with lowest samples first. Will need to change the code a bit if 
        # that isn't always the case.
        
        # beam id's can be:
        # SMSU - non-match-filtered samples intended for generating audio output
        # SMSM - main beam matched-filtered samples
        # SMSx - matched-filtered split beams, where x is:
        #   F - forward
        #   B - back
        #   S - starboard
        #   P - port
        
        # print(raw['datagram_number'], raw['sample_index'], raw['ping_number'],
        #       raw['beam_index_start'], num_samples)
        
        if raw['id'] == 'SMSM':
            # a SMSM message with no samples and the most significant bit of the
            # datagram_number field set marks the end of the current ping
            if (raw['datagram_number'] & (1 << 31)) != 0:
                # But we use the EOP datagram rather than this empty RAW2 datagram to 
                # know when to process the backscatter into Sv and TS
                return

            # For the horizontal beams it looks like beam_index_start is 0, otherwise
            # the datagram's data is from a different beam/fan
            if raw['beam_index_start'] == 0:
                # don't need the complex values so save some space...
                self._power.append(20.0*np.log10(np.abs(raw['data'])))


    def _extract_ping_config(self, pco: dict) -> None:
        """Extract various parameters that are needed to calculate Sv and TS."""
        
        # TODO assumes that there is only one transcevier config!!!
        cfg =  pco['ping_configuration']['transceiver_config'][0]

        # Find which fan dataset is self.selected_fan_name
        rx_fan = [fan for fan in cfg['rx_config']['fans']
                  if fan['fan_name'] == self.selected_fan_name]
        # and same for selected ping
        tx_config = [ping for ping in cfg['tx_config']['tx_ping_config']
                     if ping['ping_name'] == self.selected_ping_name]
        
        if not rx_fan:
            logger.error('No fan with name of %s found in the ping configuration datagram',
                         self.selected_fan_name)

        # for less typing below
        rx_fan = rx_fan[0]
        tx_config = tx_config[0]

        self.sample_interval = rx_fan['sample_interval']
        # there is an absorption value for each beam so use that instead of this one
        # self.absorption_coefficient = rx_fan['rx_beams'][0]['performance_info']['absorption_coefficient']
        self.transmit_power = tx_config['performance_info']['tx_power']
        self.frequency = tx_config['frequency']
        self.pulse_duration = tx_config['pulse_duration']
        
        g = []
        g_a = []
        sa = []
        sa_a = []
        eba = []
        alpha = []
        th = []
        tl = []
        lbls = []

        beams = rx_fan['rx_beams']

        for b in beams:
            lbls.append(b['beam_name'][-3:]) # pick out the number part of the beam name

            _, inc, azi = cartesian_to_spherical(b['steering_vector_hcs_x'],
                                                 b['steering_vector_hcs_y'],
                                                 b['steering_vector_hcs_z'])
            th.append(azi)
            tl.append(inc)

            g.append(b['performance_info']['gain'])
            g_a.append(b['performance_info']['gain_adjust'])
            sa.append(b['performance_info']['sa_correction'])
            sa_a.append(b['performance_info']['sa_correction_adjust'])
            alpha.append(b['performance_info']['absorption_coefficient'])

            EBA = b['performance_info']['equivalent_beam_angle']
            if EBA == 0.0:
                # Use the classic Simrad approximate formula
                EBA = 10*np.log10(b['beam_width_x'] * b['beam_width_y'] / 5800)
                logger.warning('No EBA for beam %s - using %.1f dB', b['beam_name'], EBA)
            eba.append(EBA)


        # TODO - check that the direction of these angles is as required by the polar plot
        # - first glance suggests it is not!
        
        self.theta = np.array(th)
        self.tilts = np.pi/2 - np.array(tl)
        self.labels = np.array(lbls)
        self.gain_rx = np.array(g)
        self.gain_adjust = np.array(g_a)
        self.sa_correction = np.array(sa)
        self.sa_correction_adjust = np.array(sa_a)
        self.absorption_coefficient = np.array(alpha).mean()  # same for all beams one expects!
        self.equivalent_beam_angle = np.array(eba)



"""Manage and provide access to the sonarcal config file."""

from pathlib import Path
import configparser
from platformdirs import PlatformDirs


def int_config(func):
    """A decorator to set and return an int from the configuration."""
    def wrapper(self, value=None):
        n = func(self, value)
        if value:
            self.sc[n] = str(value)
        return self.sc.getint(n)
    return wrapper

def float_config(func):
    """A decorator to set and return a float from the configuration."""
    def wrapper(self, value=None):
        n = func(self, value)
        if value:
            self.sc[n] = str(value)
        return self.sc.getfloat(n)
    return wrapper

def bool_config(func):
    """A decorator to set and return a boolean from the configuration."""
    def wrapper(self, value=None):
        n = func(self, value)
        if value is not None:
            self.sc[n] = str(value)
        return self.sc.getboolean(n)
    return wrapper

def str_config(func):
    """A decorator to set and return a str from the configuration."""
    def wrapper(self, value=None):
        n = func(self, value)
        if value:
            self.sc[n] = value
        return self.sc[n]
    return wrapper

def pathlib_config(func):
    """A decorator to set and return a pathlib Path from the configuration."""
    def wrapper(self, value=None):
        n = func(self, value)
        if value:
            self.sc[n] = str(value)
        return Path(self.sc[n])
    return wrapper


class sonarcalConfig():

    def __init__(self):
        self.app_name = 'sonarcal'
        self.app_author = 'Aqualyd'
        self.ini_section_name = 'sonarcal'
        
        self.dirs = PlatformDirs(appname=self.app_name, appauthor=self.app_author)

        self.config_filename = Path(self.dirs.user_config_dir)/'config.ini'
        self.config_filename.parent.mkdir(parents=True, exist_ok=True)

        self.config = configparser.ConfigParser()
        c = self.config.read(self.config_filename, encoding='utf8')

        # default values, used if an entry is not present in the config file or 
        # the config file doesn't exist.
        defaults = {'title': 'Sonar calibration',
                    'spherets': '-35.0',
                    'numPingsToShow': '100',
                    'maxRange': '50',
                    'maxSv': '-20',
                    'minSv': '-60',
                    'realtimeReplay': 'yes',
                    'replayPingInterval': '0.2',
                    'horizontalBeamGroupPath': 'Sonar/Beam_group1',
                    'watchDir': '.',
                    'liveData': 'no',
                    'calibrating_colour': '#EE9A00',  # an orange
                    'movingAveragePoints': '10',
                    'sphereStatsOver': '5',
                    'sliderLowestSv': '-100',
                    'sliderHighestSv': '10',
                    'diffPlotYMin': '-3.0'
                    }

        if not c:  # config file not found, so make one
            self.config[self.ini_section_name] = defaults
            self.save_config()
        else:
            # if the config file doesn't have all of the entries in the defaults, add 
            # the missing ones in.
            if not self.config.has_section(self.ini_section_name):
                self.config.add_section(self.ini_section_name)
            added = False
            for k, v in defaults.items():
                if k not in self.config[self.ini_section_name]:
                    added = True
                    self.config[self.ini_section_name][k] = v
            if added:
                self.save_config()

        self.sc = self.config[self.ini_section_name]

    # Note: decorated functions below can get and set config values.
    # Non-decorated functions only get config values.
    def save_config(self):
        with open(self.config_filename, 'w', encoding='utf-8') as configfile:
            self.config.write(configfile)
    
    def appName(self) -> str:
        return self.app_name

    def appAuthor(self) -> str:
        return self.app_author

    @float_config
    def sphereTS(self, value=None):
        return 'sphereTS'

    @staticmethod
    def title() -> str:
        # Used for the main window title bar text
        return 'Sonar calibration'

    @staticmethod
    def helpURI() -> Path:
        return Path(__file__).parent/'offline-docs'/'index.html'
    
    @staticmethod
    def iconFile() -> Path:
        return Path(__file__).parent/'assets'/'logo.png'
    
    def autoSaveDir(self) -> Path:
        d = Path(self.dirs.user_data_dir)/'autosave'
        d.mkdir(parents=True, exist_ok=True)
        return d

    def logDir(self) -> Path:
        d = Path(self.dirs.user_log_dir)
        d.mkdir(parents=True, exist_ok=True)
        return d
    
    def userDocumentsDir(self) -> str:
        return self.dirs.user_documents_dir
    
    @int_config
    def numPings(self, value=None):
        return 'numPingsToShow'
    
    @float_config
    def maxRange(self, value=None):
        return 'maxRange'

    @float_config  
    def maxSv(self, value=None):
        return 'maxSv'
    
    @float_config
    def minSv(self, value=None):
        return 'minSv'

    @float_config
    def diffPlotYMin(self, value=None):
        return 'diffPlotYMin'

    @float_config
    def replayPingInterval(self, value=None):
        return 'replayPingInterval'

    @str_config
    def horizontalBeamGroup(self, value=None):
        return 'horizontalBeamGroupPath'

    @pathlib_config
    def watchDir(self, value=None) -> str:
        return 'watchDir'

    @bool_config
    def liveData(self, value=None):
        return 'liveData'

    @str_config
    def calibrating_colour(self, value=None):
        # Colour used for highlighting things when calibrating a beam
        return 'calibrating_colour'

    @int_config
    def sphereStatsOver(self, value=None):
        """Number of sphere values to use for the ping-to-ping variability."""
        return 'sphereStatsOver'

    @int_config
    def movingAveragePoints(self, value=None):
        """Number of points for moving average for smoothed plots."""
        return 'movingAveragePoints'

    @float_config
    def sliderLowestSv(self, value=None):
        "Echogram colour scheme lower threshold."
        return 'sliderLowestSv'
        
    @float_config
    def sliderHighestSv(self, value=None):
        "Echogram colour scheme upper threshold."
        return 'sliderHighestSv'

# Simple way to get a singleton...
config = sonarcalConfig()



import sys
from time import sleep
# import h5py
import numpy as np
from datetime import datetime
from .utils import beamAnglesFromNetCDF4, SvTSFromSonarNetCDF4, nt_time_to_datetime
from .datagram_processor import rawDatagramProcessor
import logging
from pathlib import Path
from .configuration import config
from .raw_parser import simrad_raw_file as raw


logger = logging.getLogger(config.appName())

def most_recent_file(watch_dir: Path, wait_interval: float=5.0):
    """Get the most recent .nc or .raw file in the directory."""

    while True:
        files = sorted(list(watch_dir.glob('*.nc')) + list(watch_dir.glob('*.raw')))

        if files:
            return files[-1]

        logger.info("No .nc or .raw files found in '%s'", watch_dir)
        sleep(wait_interval)


def file_type(filename: Path):
    """Works out what sonar the data file is from and what format it is."""
    
    match filename.suffix:
        case '.nc':
            return 'sonar-netcdf4'
        case '.raw':
            return 'simrad-raw'
    return ''


def sonar_file_read(msg_queue):
    """Run code to listen to or read from the last file in the watched directory."""
    
    watch_dir = config.watchDir()
    beam_group = config.horizontalBeamGroup()
    live_data = config.liveData()

    last_file = most_recent_file(watch_dir)

    # beamGroup is currently only correct for netcdf files, not raw files.
    
    match file_type(last_file):
        case 'sonar-netcdf4':
            if live_data:
                file_listen_netcdf(watch_dir, beam_group, msg_queue)
            else:
                file_replay_netcdf(watch_dir, beam_group, msg_queue)
        case 'simrad-raw':
            if live_data:
                file_listen_raw(watch_dir, msg_queue)
            else:
                file_replay_raw(watch_dir, msg_queue)
        case _:
            logger.error('Unsupported sonar file type')


def file_listen_netcdf(watchDir, beamGroup, msg_queue):
    """Listen for new data in a file.

    Find new data in the most recent file (and keep checking for more new data).
    Used for live calibrations.
    """
    # A more elegant method for all of this can be found in the examples here:
    # https://docs.h5py.org/en/stable/swmr.html, which uses the watch facility
    # in the hdf5 library (but we're not sure if the omnisonars write data in
    # a manner that this will work with).

    # Config how and when to give up looking for new data in an existing file.
    maxNoNewDataCount = 20  # number of tries to find new pings in an existing file
    waitInterval = 0.5  # [s] time period between checking for new pings
    waitIntervalFile = 1.0  # [s] time period between checking for new files
    errorWaitInterval = 0.2  # [s] time period to wait if there is a file read error

    pingIndex = -1  # which ping to read. -1 means the last ping, -2 the second to last ping

    t_previous = datetime(1970, 1, 1)  # timestamp of previous ping
    f_previous = ''  # previously used file

    while True:  # could add a timeout on this loop...
        mostRecentFile = most_recent_file(watchDir, waitIntervalFile)

        if mostRecentFile == f_previous:  # no new file was found
            logger.info('No newer file found. Will try again in %s s.', str(waitIntervalFile))
            sleep(waitIntervalFile)  # wait and try again
        else:
            logger.info('Listening to file: %s.', mostRecentFile)
            noNewDataCount = 0

            while noNewDataCount <= maxNoNewDataCount:
                # open netcdf file
                try:
                    import h5py  # deferred to save startup time
                    f = h5py.File(mostRecentFile, 'r', libver='latest', swmr=True)
                    # f = h5py.File(mostRecentFile, 'r') # without HDF5 swmr option
                    f_previous = mostRecentFile

                    t = nt_time_to_datetime(f[beamGroup + '/ping_time'][pingIndex]/100)

                    if t > t_previous:  # there is a new ping in the file

                        theta, tilts = beamAnglesFromNetCDF4(f, beamGroup, pingIndex)
                        sv, ts, gains = SvTSFromSonarNetCDF4(f, beamGroup, pingIndex, tilts)

                        samInt = f[beamGroup + '/sample_interval'][pingIndex]
                        c = f['Environment/sound_speed_indicative'][()]
                        labels = f[beamGroup + '/beam']

                        # convert HDF5 text to list of str
                        labels = np.array([s.decode('utf-8') for s in labels])

                        t_previous = t
                        noNewDataCount = 0  # reset the count
                       
                        # send the data off to be plotted
                        msg_queue.put((t, samInt, c, sv, ts, theta, tilts, gains, labels))
                    else:
                        noNewDataCount += 1
                        if noNewDataCount > maxNoNewDataCount:
                            logger.info('No new data found in file %s after waiting %.1f s.',
                                         mostRecentFile.name, noNewDataCount * waitInterval)

                    f.close()
                    # try this instead of opening and closing the file
                    # t.id.refresh(), etc
                    sleep(waitInterval)
                except OSError:
                    f.close()  # just in case...
                    e = sys.exc_info()
                    logger.warning('OSError when reading netCDF4 file:')
                    logger.warning(e)
                    logger.warning('Ignoring the above and trying again.')
                    sleep(errorWaitInterval)


def file_replay_netcdf(watchDir, beamGroup, msg_queue):
    """Replay all data in the directory. Used for testing."""

    files = list(watchDir.glob('*.nc'))

    if not files:
        logger.info('No .nc files in %s', watchDir)

    for file in sorted(files, key=lambda p: p.stem):
        logger.info('Replaying file: %s', file)

        # open netcdf file
        import h5py  # deferred to save startup time
        f = h5py.File(file, 'r')

        t = f[beamGroup + '/ping_time']

        # Send off each ping at a sedate rate...
        for i in range(0, t.shape[0]):
            theta, tilts = beamAnglesFromNetCDF4(f, beamGroup, i)
            sv, ts, gains = SvTSFromSonarNetCDF4(f, beamGroup, i, tilts)

            samInt = f[beamGroup + '/sample_interval'][i]
            c = f['Environment/sound_speed_indicative'][()]
            labels = f[beamGroup + '/beam']

            # convert HDF5 text to list of str
            labels = np.array([s.decode('utf-8') for s in labels])

            # send the data off to be plotted
            ping_time = nt_time_to_datetime(t[i]/100)
            msg_queue.put((ping_time, samInt, c, sv, ts, theta, tilts, gains, labels))

            sleep(config.replayPingInterval())

        f.close()
    
    logger.info('Finished replaying files in %s', watchDir)


def file_listen_raw(watchDir: Path, msg_queue):
    """Replay live files."""

    previous_file = None

    # Check this often for files to appear in the directory if there are none
    file_wait = 2.0  # [s]
    
    # Check this often for new files to appear in the directory once we've finished
    # reading an existing file
    new_file_wait = 2.0  # [s]

    while True:
        files = list(watchDir.glob('*.raw'))

        if not files:
            logger.info('No .raw files in %s. Waiting...', watchDir)
            sleep(file_wait)
            continue

        sorted_files = sorted(files, key=lambda p: p.stem)
        last_file = sorted_files[-1]

        if last_file == previous_file:
            # there is no new file, we've already read through the
            # most recent file, so perhaps the sonar has finished
            # recording. We'll wait for more...
            logger.info('No new .raw files to read. Waiting for more...')
            sleep(new_file_wait)
            continue

        previous_file = last_file

        # there is a new raw file to read
        with raw.RawSimradFile(last_file) as fid:
            proc = rawDatagramProcessor()
            # read and process datagrams in last_file as they get written to the file
            logger.info('Reading datagrams from %s', last_file.name)

            while True:
                # live_read() will block waiting for a new datagram to be appended to the file.
                # If nothing gets appended after a few seconds it raises a
                # SimradFileFinished exception
                try:
                    dg = fid.live_read()
                    
                    if proc.add_datagram(dg):  # returns True when a processed ping is available
                        
                        msg_queue.put((proc.ping_time, proc.sample_interval, proc.sound_speed,
                                      proc.sv, proc.ts, proc.theta, proc.tilts,
                                      proc.gain_rx, proc.labels))

                        sleep(config.replayPingInterval())
                except raw.SimradFileFinished:
                    break  # go back to the outer 'while True' loop to look for a new file.


def file_replay_raw(watchDir: Path, msg_queue):
    """Replays raw files."""

    files = list(watchDir.glob('*.raw'))

    if not files:
        logger.info('No .raw files in %s', watchDir)

    for file in sorted(files, key=lambda p: p.stem):
        with raw.RawSimradFile(file) as fid:
            proc = rawDatagramProcessor()
            logger.info('Reading datagrams from %s', file.name)

            while True:
                try:
                    dg = fid.read(1)

                    if proc.add_datagram(dg):  # returns True when a processed ping is available
                        msg_queue.put((proc.ping_time, proc.sample_interval, proc.sound_speed,
                                      proc.sv, proc.ts, proc.theta, proc.tilts, proc.gain_rx,
                                      proc.labels))
                        sleep(config.replayPingInterval())
                except raw.SimradEOF:
                    break  # go back to the outer 'while True' loop for the next file
    
    logger.info('Finished replaying files in %s', watchDir)



import traceback
from queue import Empty
from datetime import datetime, timezone
import matplotlib.pyplot as plt
import matplotlib as mpl
import numpy as np
# from scipy import signal
from matplotlib.widgets import RangeSlider
from .gui_utils import draggable_ring, draggable_radial
import humanize
import logging
from .configuration import config

# Matplotlib for tkinter
mpl.use('TkAgg')

logger = logging.getLogger(config.appName())

class echogramPlotter:
    """Receive via a queue new ping data and use that to update the display."""

    def __init__(self, msg_queue, root):

        # This flag changes to True once self.createGUI() has run
        self.gui_created = False

        self.queue = msg_queue
        self.root = root
        self.job = None

        # All callback that is called whenever a new ping has finished drawing
        self.new_ping_cb = None

        # Various user-changable lines on the plots that could in the future
        # come from a config file.
        self.beamLine = None
        self.beamLineAngle = 0.0  # [deg]
        self.beamIdx = 0  # dummy value. Is updated once some data are received.
        self.beamLabel = ''

        self.minTargetRange = 0.33*config.maxRange()
        self.maxTargetRange = 0.66*config.maxRange()

        self.varNum = config.sphereStatsOver()

        self.diffPlotYlim = (config.diffPlotYMin(), 0)  # [dB]

        self.numPings = config.numPings()  # to show in the echograms
        self.maxRange = config.maxRange()  # [m] of the echograms
        self.maxSv = config.maxSv()  # [dB] max Sv to show in the echograms
        self.minSv = config.minSv()  # [dB] min Sv to show in the echograms

        self.checkQueueInterval = 200  # [ms] duration between checking the queue for new data

        self.movingAveragePoints = config.movingAveragePoints()

        self.emptySv = -999.0  # initialisation value of echogram data

        # Make the plots. It gets filled with pretty things once the first ping
        # of data is received.
        self.fig = plt.figure(figsize=(11.5, 5))
        plt.ion()

        self.firstPing = True

    def createGUI(self, samInt, c, sv, ts, theta, tilts, labels):
        """Create the GUI."""
        cmap = mpl.colormaps['jet']  # viridis looks nice too...
        cmap.set_under('w')  # and for values below self.minSv, if desired

        # number of samples to store per ping
        self.maxSamples = int(np.ceil(self.maxRange / (samInt*c/2.0)))
        self.numBeams = sv.shape[0]
        
        # A copy of the beam gains from the file being read
        self.gains = None

        # Storage for the things we plot
        # Polar plot
        self.polar = np.ones((self.maxSamples, self.numBeams), dtype=float) * self.emptySv
        # Echograms
        self.port = np.ones((self.maxSamples, self.numPings), dtype=float) * self.emptySv
        self.main = np.ones((self.maxSamples, self.numPings), dtype=float) * self.emptySv
        self.stbd = np.ones((self.maxSamples, self.numPings), dtype=float) * self.emptySv
        # Amplitude of sphere
        self.amp = np.ones((3, self.numPings), dtype=float) * np.nan
        self.ampSmooth = np.ones((3, self.numPings), dtype=float) * np.nan
        # Range of the max amplitude within the range range on the selected beam
        self.rangeMax = None  # [m]

        # Differences in sphere amplitudes, smoothed version
        self.ampDiffPort = np.ones((self.numPings), dtype=float) * np.nan
        self.ampDiffStbd = np.ones((self.numPings), dtype=float) * np.nan
        self.ampDiffPortSmooth = np.ones((self.numPings), dtype=float) * np.nan
        self.ampDiffStbdSmooth = np.ones((self.numPings), dtype=float) * np.nan

        # Make the plot axes and set up static things
        self.polarPlotAx = plt.subplot2grid((3, 3), (0, 0), rowspan=3, projection='polar')
        self.portEchogramAx = plt.subplot2grid((3, 3), (0, 1))
        self.mainEchogramAx = plt.subplot2grid((3, 3), (1, 1))
        self.stbdEchogramAx = plt.subplot2grid((3, 3), (2, 1))
        self.ampPlotAx = plt.subplot2grid((3, 3), (0, 2), rowspan=2)
        self.ampDiffPlotAx = plt.subplot2grid((3, 3), (2, 2))

        # plt.tight_layout(pad=1.5, w_pad=0.0, h_pad=0.0)

        # Configure the echogram axes
        self.portEchogramAx.invert_yaxis()
        self.mainEchogramAx.invert_yaxis()
        self.stbdEchogramAx.invert_yaxis()

        self.portEchogramAx.yaxis.tick_right()
        self.mainEchogramAx.yaxis.tick_right()
        self.stbdEchogramAx.yaxis.tick_right()

        self.portEchogramAx.xaxis.set_ticklabels([])
        self.mainEchogramAx.xaxis.set_ticklabels([])

        # Configure the sphere amplitude axes
        self.ampPlotAx.yaxis.tick_right()
        self.ampPlotAx.yaxis.set_label_position("right")
        self.ampPlotAx.xaxis.set_ticklabels([])
        self.ampPlotAx.grid(axis='y', linestyle=':')
        self.ampDiffPlotAx.yaxis.tick_right()
        self.ampDiffPlotAx.yaxis.set_label_position("right")
        self.ampDiffPlotAx.grid(axis='y', linestyle=':')

        self.portEchogramAx.set_title('', loc='left', color='red')
        self.mainEchogramAx.set_title(f' {self.beamLabel}', loc='left')
        self.stbdEchogramAx.set_title('', loc='left', color='green')

        # Create the lines in the plots
        # Sphere TS from 3 beams
        self.ampPlotLinePort, = self.ampPlotAx.plot(self.amp[0, :], 'r-', linewidth=1)
        self.ampPlotLineMain, = self.ampPlotAx.plot(self.amp[1, :], 'k-', linewidth=1)
        self.ampPlotLineStbd, = self.ampPlotAx.plot(self.amp[2, :], 'g-', linewidth=1)
     
        # Smoothed curves for the TS from 3 beams
        self.ampPlotLinePortSmooth, = self.ampPlotAx.plot(self.ampSmooth[0, :], 'r-', linewidth=2)
        self.ampPlotLineMainSmooth, = self.ampPlotAx.plot(self.ampSmooth[1, :], 'k-', linewidth=2)
        self.ampPlotLineStbdSmooth, = self.ampPlotAx.plot(self.ampSmooth[2, :], 'g-', linewidth=2)
        self.ampPlotAx.set_xlim(0, self.numPings)
     
        # a informative number on the TS plot
        self.diffVariability = self.ampPlotAx.text(0.05, 0.95, '', ha='left', va='top',
                                                   transform=self.ampPlotAx.transAxes)
        self.diffVariability.set_bbox({'color': 'w', 'alpha': 0.5})

        # Difference in sphere TS from the 3 beams
        self.ampDiffPortPlot, = self.ampDiffPlotAx.plot(self.ampDiffPort, 'r-', linewidth=1)
        self.ampDiffStbdPlot, = self.ampDiffPlotAx.plot(self.ampDiffStbd, 'g-', linewidth=1)
        # Smoothed curves of the difference in TS
        self.ampDiffPortPlotSmooth, = self.ampDiffPlotAx.plot(self.ampDiffPortSmooth, 'r-',
                                                              linewidth=2)
        self.ampDiffStbdPlotSmooth, = self.ampDiffPlotAx.plot(self.ampDiffStbdSmooth, 'g-',
                                                              linewidth=2)
        self.ampDiffPlotAx.set_xlim(0, self.numPings)
        self.ampDiffPlotAx.set_ylim(self.diffPlotYlim)

        # Echograms for the 3 selected beams
        ee = [0.0, self.numPings, self.maxRange, 0.0]
        self.portEchogram = self.portEchogramAx.imshow(self.port, interpolation='nearest',
                                                       aspect='auto', extent=ee, vmin=self.minSv,
                                                       vmax=self.maxSv)
        self.mainEchogram = self.mainEchogramAx.imshow(self.main, interpolation='nearest',
                                                       aspect='auto', extent=ee, vmin=self.minSv,
                                                       vmax=self.maxSv)
        self.stbdEchogram = self.stbdEchogramAx.imshow(self.stbd, interpolation='nearest',
                                                       aspect='auto', extent=ee, vmin=self.minSv,
                                                       vmax=self.maxSv)

        self.portEchogram.set_cmap(cmap)
        self.mainEchogram.set_cmap(cmap)
        self.stbdEchogram.set_cmap(cmap)

        # Omni echogram axes setup
        self.polarPlotAx.set_theta_offset(np.pi/2)  # to make bow direction plot upwards
        self.polarPlotAx.set_frame_on(False)
        self.polarPlotAx.xaxis.set_ticklabels([])

        # Omni echogram image
        r = np.arange(0, self.maxSamples)*samInt*c/2.0
        self.polarPlot = self.polarPlotAx.pcolormesh(theta, r, self.polar,
                                                     shading='auto', vmin=self.minSv,
                                                     vmax=self.maxSv)
        self.polarPlotAx.grid(axis='y', linestyle=':')

        self.polarPlot.set_cmap(cmap)

        # Colorbar for the omni echogram
        cb = plt.colorbar(self.polarPlot, ax=self.polarPlotAx, orientation='horizontal',
                          extend='both', fraction=0.05, location='bottom')
        cb.set_label('$S_v$ re 1 m$^{-1}$ [dB]')

        # Range rings on the omni echogram
        self.rangeRing1 = draggable_ring(self.polarPlotAx, self.minTargetRange)
        self.rangeRing2 = draggable_ring(self.polarPlotAx, self.maxTargetRange)
        self.beamLine = draggable_radial(self.polarPlotAx, self.beamLineAngle,
                                         self.maxRange, theta, labels)

        # sets self.beamIdx and self.beamLabel from the positon of the radial line
        self.updateBeamNum(theta)  

        # Axes labels
        self.stbdEchogramAx.set_xlabel('Pings')

        self.portEchogramAx.yaxis.set_label_position('right')

        self.mainEchogramAx.yaxis.set_label_position('right')
        self.mainEchogramAx.set_ylabel('Range (m)')

        self.stbdEchogramAx.yaxis.set_label_position('right')

        self.ampDiffPlotAx.set_xlabel('Pings')
        self.ampPlotAx.set_ylabel('TS re 1 m$^{2}$ [dB]')
        self.ampDiffPlotAx.set_ylabel(r'$\Delta$ (dB)')
        self.ampPlotAx.set_title('Maximum amplitude at 0 m')

        plt.tight_layout(pad=1.5, w_pad=0.0, h_pad=0.0)
        
        # the tight_layout call causes a resize of the polar plot, so update the
        # inverse transform that beamline keeps
        self.beamLine.resized(None)
        
        # Text for the horizontal beam tilt
        self.beamTiltText = self.polarPlotAx.text(0.05, 1.0, s='',
                                                  transform=self.polarPlotAx.transAxes)

        # range slider to adjust the echogram thresholds. Do this after the tight_layout
        # call as it otherwise complains
        slider_ax = plt.axes([0.006, 0.2, 0.015, 0.6])
        lowestSv = config.sliderLowestSv()
        highestSv = config.sliderHighestSv()

        self.slider = RangeSlider(slider_ax, label="Thresholds", valmin=lowestSv, valmax=highestSv,
                                  valinit=((self.minSv, self.maxSv)),
                                  valstep=1.0,
                                  orientation='vertical', facecolor='blue')
        # self.slider.valtext.set_rotation(90)
        self.slider.valtext.set_visible(False)
        self.slider.label.set_rotation(90)
        self.slider.on_changed(self.updateEchogramThresholds)
        
        self.gui_created = True
        
        self.updateTiltValue(tilts.mean())

    def updateEchogramThresholds(self, val):
        """Update the image colormaps."""
        self.polarPlot.set_clim(val)
        self.portEchogram.set_clim(val)
        self.mainEchogram.set_clim(val)
        self.stbdEchogram.set_clim(val)

        # Redraw the figure to ensure it updates
        self.fig.canvas.draw_idle()

    def updateTiltValue(self, tilt: float):
        """Update the display of beam tilt on GUI.
        
        Units of tilt are radians
        """
        if self.gui_created:
            self.beamTiltText.set_text(f'Tilt: {tilt*180/np.pi:0.1f}°')

    def updateDiffPlotYLim(self):
        """Update the lower limit of the difference plot y-axis."""
        if self.gui_created:
            self.ampDiffPlotAx.set_ylim((config.diffPlotYMin(), 0))

    def updateRangeSliderSettings(self):
        """Update the echogram range slider min and max values from the config."""
        if self.gui_created:
            vmin, vmax = self.slider.val
            if vmin < config.sliderLowestSv():
                self.slider.set_min(config.sliderLowestSv())
                config.minSv(config.sliderLowestSv())  # TODO - doesn't update the config dialog...
            if vmax > config.sliderHighestSv():
                self.slider.set_max(config.sliderHighestSv())
                config.maxSv(config.sliderHighestSv())  # TODO - doesn't update the config dialog...

            self.slider.valmin = config.sliderLowestSv()
            self.slider.valmax = config.sliderHighestSv()
            self.slider.ax.set_ylim(self.slider.valmin, self.slider.valmax)

        self.fig.canvas.draw_idle()

    def set_ping_callback(self, cb):
        """Set the callback that is called after each new ping is displayed."""
        self.new_ping_cb = cb

    @staticmethod
    def sort_beams(sv, ts, theta, tilts, gains, labels):
        """Sorts everything by the values in theta."""
        
        # sort by the negative of theta because of Matplotlib's polar plot has increasing
        # angle values to anti-clockwise, but Sonarcal has then to clockwise.
        sort_i = np.argsort(-theta)

        return sv[sort_i], ts[sort_i], theta[sort_i], tilts[sort_i], gains[sort_i], labels[sort_i]

    def newPing(self, label):
        """Receive messages from the queue, decodes them and updates the echogram."""
        while not self.queue.empty():
            try:
                message = self.queue.get(block=False)
            except Empty:
                logger.info('No new data in received message.')
            else:
                try:
                    (pingTime, samInt, c, sv, ts, theta, tilts, gains, labels) = message
                    # pingTime - datetime
                    # samInt - sample interval: float [m]
                    # c - sound speed: float [m/s]
                    # sv - Sv: 2D numpy float [dB]
                    # ts - TS: 2D numpy float [dB]
                    # the rest are all 1d numpy vectors
                    # theta - float [rad]
                    # tilts - float [rad]
                    # gains - float [dB]
                    # labels - str 

                    # sort on theta - needed to avoid a warning from the polar plot
                    (sv, ts, theta, tilts, gains, labels) =\
                        self.sort_beams(sv, ts, theta, tilts, gains, labels)

                    if self.firstPing:
                        self.firstPing = False
                        self.createGUI(samInt, c, sv, ts, theta, tilts, labels)

                    # Update our copy of the beam gains
                    self.gains = gains

                    # Update the plots with the data in the new ping
                    timeBehind = datetime.now(timezone.utc) - pingTime
                    milliseconds = pingTime.microsecond / 1000
                    label.config(text=f'Ping at {pingTime:%Y-%m-%d %H:%M:%S}.' +
                                 f'{milliseconds:03.0f} '
                                 f'({humanize.precisedelta(timeBehind)} ago)')
                    logger.info('Displaying ping from %s.', pingTime)

                    self.minTargetRange = min(self.rangeRing1.range, self.rangeRing2.range)
                    self.maxTargetRange = max(self.rangeRing1.range, self.rangeRing2.range)

                    # print('Range rings: {}, {}'.format(self.minTargetRange, self.maxTargetRange))

                    minSample = int(np.floor(2*self.minTargetRange / (samInt * c)))
                    maxSample = int(np.floor(2*self.maxTargetRange / (samInt * c)))

                    self.updateBeamNum(theta)  # sets self.beam from self.beamLineAngle

                    # work out the beam indices
                    if self.beamIdx == 0:
                        beamPort = self.numBeams-1
                    else:
                        beamPort = self.beamIdx-1

                    if self.beamIdx == self.numBeams-1:
                        beamStbd = 0
                    else:
                        beamStbd = self.beamIdx+1

                    # Find the max ts between the min and max ranges set by the UI
                    # and store for plotting
                    self.amp = np.roll(self.amp, -1, 1)
                    self.amp[0, -1] = np.max(ts[beamPort][minSample:maxSample])
                    max_i = np.argmax(ts[self.beamIdx][minSample:maxSample])
                    self.amp[1, -1] = ts[self.beamIdx][minSample+max_i]
                    self.rangeMax = (minSample+max_i) * samInt * c / 2.0
                    self.amp[2, -1] = np.max(ts[beamStbd][minSample:maxSample])

                    # Store the amplitude for the 3 beams for the echograms
                    self.port = self.updateEchogramData(self.port, sv[beamPort])
                    self.main = self.updateEchogramData(self.main, sv[self.beamIdx])
                    self.stbd = self.updateEchogramData(self.stbd, sv[beamStbd])

                    # Update the plots
                    # Sphere TS from 3 beams
                    self.ampPlotLinePort.set_ydata(self.amp[0, :])
                    self.ampPlotLineMain.set_ydata(self.amp[1, :])
                    self.ampPlotLineStbd.set_ydata(self.amp[2, :])
                    # and smoothed plots
                    coeff = np.ones(self.movingAveragePoints)/self.movingAveragePoints
                    # and measure of ping-to-ping variability
                    variability = np.std(self.amp[1, -self.varNum: -1])
                    if not np.isnan(variability):
                        self.diffVariability.set_text(rf'$\sigma$ = {variability:.1f} dB')

                    from scipy import signal  # deferred to save startup time
                    self.ampSmooth[0, :] = signal.filtfilt(coeff, 1, self.amp[0, :])
                    self.ampSmooth[1, :] = signal.filtfilt(coeff, 1, self.amp[1, :])
                    self.ampSmooth[2, :] = signal.filtfilt(coeff, 1, self.amp[2, :])
                    self.ampPlotLinePortSmooth.set_ydata(self.ampSmooth[0, :])
                    self.ampPlotLineMainSmooth.set_ydata(self.ampSmooth[1, :])
                    self.ampPlotLineStbdSmooth.set_ydata(self.ampSmooth[2, :])

                    self.ampPlotAx.set_title(f'Maximum amplitude at {self.rangeMax:.1f} m')
                    self.ampPlotAx.relim()
                    self.ampPlotAx.autoscale_view()

                    # Difference in sphere TS from 3 beams
                    diffPort = self.amp[0, :] - self.amp[1, :]
                    diffStbd = self.amp[2, :] - self.amp[1, :]
                    self.ampDiffPortPlot.set_ydata(diffPort)
                    self.ampDiffStbdPlot.set_ydata(diffStbd)
                    # and the smoothed
                    smPort = signal.filtfilt(coeff, 1, diffPort)
                    smStbd = signal.filtfilt(coeff, 1, diffStbd)
                    self.ampDiffPortPlotSmooth.set_ydata(smPort)
                    self.ampDiffStbdPlotSmooth.set_ydata(smStbd)

                    self.ampDiffPlotAx.relim()
                    self.ampDiffPlotAx.autoscale_view(scaley=False)

                    # Beam echograms
                    self.portEchogram.set_data(self.port)
                    self.mainEchogram.set_data(self.main)
                    self.stbdEchogram.set_data(self.stbd)

                    self.portEchogramAx.set_title(f'Beam {labels[beamPort]}', loc='left')
                    self.mainEchogramAx.set_title(f'Beam {labels[self.beamIdx]}', loc='left')
                    self.stbdEchogramAx.set_title(f'Beam {labels[beamStbd]}', loc='left')

                    # Polar plot
                    for i, b in enumerate(sv):
                        if b.shape[0] > self.maxSamples:
                            self.polar[:, i] = b[0: self.maxSamples]
                        else:
                            samples = b.shape[0]
                            self.polar[:, i] =\
                                np.concatenate((b, self.emptySv*np.ones(self.maxSamples-samples)),
                                               axis=0)

                    self.polarPlot.set_array(self.polar.ravel())

                    self.updateTiltValue(tilts[self.beamIdx])

                    if self.new_ping_cb:
                        self.new_ping_cb()

                except Exception:  # if anything goes wrong, just ignore it...
                    logger.warning('Error when processing and displaying echo data:')
                    logger.warning(traceback.print_exc())
                    logger.warning('Ignoring the above and waiting for next ping.')

        self.job = self.root.after(self.checkQueueInterval, self.newPing, label)

    def updateEchogramData(self, data, pingData):
        """Shift the ping data to the left and add in the new ping data."""
        data = np.roll(data, -1, 1)
        if pingData.shape[0] > self.maxSamples:
            data[:, -1] = pingData[0:self.maxSamples]
        else:
            samples = pingData.shape[0]
            data[:, -1] = np.concatenate((pingData[:],
                                          self.emptySv*np.ones(self.maxSamples-samples)), axis=0)
        return data

    def updateBeamNum(self, theta):
        """Get the beam number from the beam line angle and the latest theta."""
        self.beamIdx = self.beamLine.selected_beam_idx
        self.beamLabel = self.beamLine.selected_beam_label



"""Re-writes Simrad raw files, datagram by datagram.

Use this script to test the sonarcal live feature, where it reads new
datagrams as they get added to a file.
"""
# /// script
# dependencies = [
#   "construct",
# ]
# ///

from time import sleep
from pathlib import Path
from construct import PaddedString, Timestamp, Bytes, Int32sl, Int64ul, this, Struct, StreamError

base_dir = Path(r'C:\Users\GavinMacaulay\OneDrive - Aqualyd Limited\Documents\Aqualyd\Projects'
                r'\2025-08 AZTI coding\example data\SN90')

raw_dir = base_dir/'Haz9'
out_dir = base_dir/'Replayed'

raw_files = sorted(raw_dir.glob('*.raw'), key=lambda p: p.stem)

# Using Construct is nice, but we could do away with it for this simple case and
# have a script with no external dependencies.

dg_def =\
    Struct(
        'size' / Int32sl,
        'type' / PaddedString(4, 'ascii'),
        'timestamp' / Timestamp(Int64ul, 1e-7, 1600),
        'data' / Bytes(this.size-12),
        'size' / Int32sl
    )

# Delete all files in out_dir
for root, dirs, files in out_dir.walk(top_down=False):
    for name in files:
        (root / name).unlink()

for in_file in raw_files:
    out_file = out_dir / in_file.name
    with open(in_file, 'rb') as fin, open(out_file, 'wb') as fout:
        print(f'Rewriting {in_file.name}')
        try:
            while True:
                dg = dg_def.parse_stream(fin)
                dg_def.build_stream(dg, fout)

                # At end of ping force a flush to disk
                if dg['type'] == 'EOP0':
                    fout.flush()
                    print(f'  End of ping at {dg["timestamp"]}')
                    sleep(1.0)

        except StreamError:
            # end of file so pause a little to simulate the sonar taking a while to start
            # a new file
            sleep(0.5)



"""Omnisonar calibration program

Provides omni and echogram displays and sphere amplitude plots for use when
calibrating sonars with horizontal beams.
"""

import tkinter as tk
from functools import partial
import threading
import queue
import sys

from .echogram_plotter import echogramPlotter
from .utils import setupLogging, on_exit, window_closed
from .file_ops import sonar_file_read
from .calibration_gui import calibrationGUI

setupLogging()

def main():
    """Omnisonar calibration graphical user interface."""    

    # queue to communicate between two threads
    msg_queue = queue.Queue()
    
    # The GUI
    root = tk.Tk()
    echogram = echogramPlotter(msg_queue, root)
    gui = calibrationGUI(echogram)

    # Sonar files are read from a separate thread
    t = threading.Thread(target=sonar_file_read, args=(msg_queue,))

    t.daemon = True  # makes the thread close when main() ends
    t.start()

    # For Windows, catch when the console is closed
    if sys.platform == "win32":
        import win32api
        win32api.SetConsoleCtrlHandler(partial(on_exit, gui.root(), gui.job()), True)

    # And start things...
    root.protocol("WM_DELETE_WINDOW", lambda: window_closed(gui.root(), gui.job()))
    root.mainloop()



# import pandas as pd
import logging
from .configuration import config

logger = logging.getLogger(config.appName())

class calibrationData():
    """Storage for sonar caliration results."""

    def __init__(self):
        import pandas as pd  # deferred to save startup time
        self.data = pd.DataFrame(columns=['Time', 'Gain (dB)', 'Target TS (dB)', 
                                          'RMS (dB)', 'Range (m)', 'Echoes'])
        self.data.index.name = 'Beam'
    
    def update(self, beam_label: str, timestamp: str, gain: float, ts: float,
               rms: float, r: float, num: int):
        self.data.loc[beam_label] = (timestamp, gain, ts, rms, r, num)
        
    def remove(self, beam_labels: list[str]):
        """Remove data for given beam."""
        self.data.drop(index=beam_labels, inplace=True)
        
    def df(self):
        return self.data  # eventually return a better form of the data?
    
    def save(self, filename:str):
        """Save the calibration data to a csv file."""

        if filename:
            logger.info('Saved results to %s', filename)

            # the dataframe index is str, but it will mostly be integer values, so make things
            # sort numerically if an int followed by non-ints.
            int_order = sorted((x for x in self.data.index if x.isdigit()), key=lambda i: int(i))
            str_order = sorted(x for x in self.data.index if not x.isdigit())
            self.data.loc[int_order+str_order].to_csv(filename)

    


# this file uses deferred imports

def calculate_gain(sphere_echoes: list[tuple], sphere_ts: float, beam_gain: float)\
    -> tuple[float, float, float, int]:
    """Calculate the beam calibration gain and other stats."""

    # deferred imports to reduce program startup time
    import pandas as pd
    import scipy.stats.mstats as ms
    import numpy as np

    df = pd.DataFrame(sphere_echoes, columns=['timestamp', 'ts', 'range'])

    # find and ignore extreme sphere echoes
    trimmed = ms.trim(df['ts'], limits=(0.05, 0.05), relative=True)
    mask = np.logical_not(np.ma.getmaskarray(trimmed))
    dfm = df[mask]

    ts_mean = 10.0 * np.log10(np.mean(np.power(10, dfm['ts']/10.0)))
    ts_rms = np.sqrt(np.mean(np.square(dfm['ts']-ts_mean)))

    # This is quite specific to the actual sonar equations - is ok for Simrad and Furuno
    # omnisonars as of 2025...

    # and the new gain correction is....
    beam_gain_new = beam_gain + ts_mean - sphere_ts

    return (beam_gain_new, ts_mean, ts_rms, np.mean(dfm['range']), len(dfm))





import webbrowser
import logging
from datetime import datetime
import tkinter as tk
from tkinter import ttk
from tkinter import messagebox
from importlib.metadata import version
from platform import python_version
from PIL import Image, ImageTk
from .utils import window_closed
from .calibration_data import calibrationData
from .calculate_gains import calculate_gain
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from .configuration import config as cfg

logger = logging.getLogger(cfg.appName())

class calibrationGUI:
    """Provides the main GUI container and misc labels/buttons."""

    def __init__(self, echogram):

        self.echogram = echogram
        self.help_uri = str(cfg.helpURI())

        # Calibration gains are stored in here
        self.cal_data = calibrationData()
        # sphere echo data for the current beam calibration is stored in here
        self.sphere_echoes = []

        # The GUI window
        self.echogram.root.title(cfg.title())
        
        # Dialogs that we keep around
        self.results_dialog = None
        
        # The toolbar and window icon/logo
        self.icon = ImageTk.PhotoImage(Image.open(cfg.iconFile()))
        self.echogram.root.iconphoto(False, self.icon)

        # Things to do with new pings 
        self.echogram.set_ping_callback(self.new_ping)

        # Put the matplotlib plots into the GUI window.
        canvas = FigureCanvasTkAgg(self.echogram.fig, master=self.echogram.root)
        canvas.get_tk_widget().pack(side='top', fill='both', expand=True, padx=5, pady=5)

        # Styles. These apply to all widgets, not just the ones created in this function
        s = ttk.Style()
        s.configure('TButton', font=('Arial', 16))
        s.configure('TLabel', font=('Arial', 12))
        s.configure('TCheckbutton', font=('Arial', 16))
        s.configure('Treeview.Heading', font=('Arial', 12, 'bold'))
        s.configure('Treeview', font=('Arial', 12))
        # this doesn't work for some reason so the font is set when the entry widget
        # is created (in the configuration dialog).
        s.configure('TEntry', font=('Arial', 20))

        # A label to show the last received message time
        self.label = ttk.Label(self.echogram.root)
        self.label.pack(side=tk.TOP, fill=tk.BOTH)
        self.label.config(text='Waiting for data...', width=100, anchor=tk.W)

        ttk.Separator(self.echogram.root, orient='horizontal').pack(fill='x', padx=10, pady=5)

        # Buttons for help, on-axis toggle, config dialog, and close
        self.onaxis_value = tk.BooleanVar(value=False)

        frame = ttk.Frame(self.echogram.root)
        results = ttk.Button(frame, text='Results', command=self.results)
        config = ttk.Button(frame, text='Config', command=self.config)
        onaxis = ttk.Checkbutton(frame, text='On-axis', variable=self.onaxis_value,
                                 command=self.onaxis_changed)
        help = ttk.Button(frame, text='Help', command=self.help)
        about = ttk.Button(frame, text='About', command=self.about)
        close = ttk.Button(frame, text='Close', command=self.close)

        onaxis.pack(side=tk.LEFT)
        close.pack(side=tk.RIGHT)
        about.pack(side=tk.RIGHT)
        help.pack(side=tk.RIGHT)
        config.pack(side=tk.RIGHT)
        results.pack(side=tk.RIGHT)

        frame.pack(side=tk.TOP, fill=tk.BOTH)

        # Start listening for sonar data
        self.echogram.newPing(self.status_label())

    def job(self):
        return self.echogram.job

    def root(self):
        return self.echogram.root

    def onaxis_changed(self):
        """A beam calibration has either started or ended."""
        if not self.echogram.beamLine:
            return

        if self.onaxis_value.get():  # start calibrating a beam
            self.echogram.beamLine.freeze(True)
            logger.info('Beam %s calibration started', self.echogram.beamLabel)
        else:  # finished calibrating a beam
            self.auto_save()
            logger.info('Beam %s calibration complete', self.echogram.beamLabel)
            self.echogram.beamLine.freeze(False)
            self.sphere_echoes = []
            if self.results_dialog:
                self.results_dialog.update_rows(None)  # unhighlights the previously active row

    def auto_save(self):
        """Save cal results to an autosave location."""
        timestamp = datetime.now().strftime('%Y%m%dT%H%M%S')
        filename = cfg.autoSaveDir()/('results_' + timestamp + '.csv')
        self.cal_data.save(filename)

    def new_ping(self):
        """Orchestrates things for each new ping."""
        e = self.echogram
        if e.beamLine.frozen():  # a beam is being calibrated
            # store the current ping's sphere echo info
            self.sphere_echoes.append((datetime.now().isoformat(), e.amp[1, -1], e.rangeMax))
            # calculate the beam gain and other stats
            (gain, ts, rms, r, num) = calculate_gain(self.sphere_echoes, cfg.sphereTS(),
                                                     e.gains[e.beamIdx])
            # store the latest beam gain values
            self.cal_data.update(e.beamLabel, datetime.now().strftime('%H:%M:%S'),
                                 gain, ts, rms, r, num)
            # update the results dialog if present
            if self.results_dialog:
                self.results_dialog.update_with(self.cal_data, e.beamLabel)

    def about(self):
        message = (f'Sonarcal, version {version("sonarcal")}, running on Python {python_version()}\n\n'
                   'Sonarcal is a program to assist with calibrating omni-directional sonars.\n\n'
                   'Developed by Aqualyd Ltd, www.aqualyd.nz')

        messagebox.showinfo(title='About', message=message)

    def close(self):
        # Catch closing the program while still calibrating a beam
        if self.echogram.beamLine and self.echogram.beamLine.frozen():
            self.auto_save()
        window_closed(self.echogram.root, self.echogram.job)

    def results(self):
        """Open the Results dialog box."""
        # want one lasting instance of this dialog so manage that here
        if not self.results_dialog:
            # deferred to reduce startup time
            from .dialog_results import resultsDialog
            self.results_dialog = resultsDialog(self.echogram.root, self.cal_data, self.icon)
        else:
            self.results_dialog.reopen()

    def help(self):
        """Open the help documentation in a web browser."""
        if not webbrowser.open(self.help_uri, new=2):
            logging.warning('Failed to start a webbrowser to show the help documentation')
        
    def config(self):
        """Open the Config dialog box."""
        from .dialog_config import configDialog  # deferred importing
        configDialog(self.echogram.root, self.icon, self.config_updated)

    def config_updated(self):
        """Things to do when the configuration is updated."""
        self.echogram.updateRangeSliderSettings()
        self.echogram.updateDiffPlotYLim()

    def status_label(self):
        return self.label



from dataclasses import dataclass
import tkinter as tk
from tkinter import ttk
from tkinter import filedialog
from .configuration import config as cfg
import logging

logger = logging.getLogger(cfg.appName())

class configDialog:
    """A dialog box to set and change application parameters."""
    def __init__(self, parent, icon=None, updated_cb=None):
        self.top = tk.Toplevel(parent)
        self.top.title("Config")
        self.updated_cb = updated_cb

        if icon:
            self.top.iconphoto(False, icon)

        config_frame = ttk.Frame(self.top)
        
        @dataclass
        class Param:
            label: str
            name: str
            type: str
            unit: str = ''
            special: str = None

        self.params = [
            Param('Settings marked with * take effect with a program restart', '', 'label'),
            Param('', '', 'horizline'),
            Param('Calibration sphere TS', 'sphereTS', 'float', 'dB re 1 m²'),
            Param('* Sonar data directory', 'watchDir', 'str', '', 'filechooser'),
            Param('* Use live data (otherwise replay)', 'liveData', 'boolean'),
            Param('Replay ping interval', 'replayPingInterval', 'float', 's'),
            Param('* Sonar-netCDF4 beam group path', 'horizontalBeamGroup', 'str'),
            Param('', '', 'horizline'),
            Param('* Plot x-axis size', 'numPings', 'int', 'pings'),
            Param('* Echogram range', 'maxRange', 'float', 'm'),
            Param('', '', 'horizline'),
            Param('Difference plot y-axis minimum', 'diffPlotYMin', 'float', 'dB'),
            Param('* Default minimum echogram Sv', 'minSv', 'float', 'dB re 1 m⁻¹'),
            Param('* Default maximum echogram Sv', 'maxSv', 'float', 'dB re 1 m⁻¹'),
            Param('Minimum allowed Sv colour', 'sliderLowestSv', 'float', 'dB re 1 m⁻¹'),
            Param('Maximum allowed Sv colour', 'sliderHighestSv', 'float', 'dB re 1 m⁻¹'),
            Param('', '', 'horizline'),
            Param('* TS smoothing over', 'movingAveragePoints', 'int', 'pings'),
            Param('* Sphere stats over', 'sphereStatsOver', 'int', 'pings'),
        ]

        self.vars = {}  # mapping for name to tkinter Var
        for p in self.params:
            if p.type == 'horizline':
                ttk.Separator(self.top, orient=tk.HORIZONTAL).pack(fill=tk.X, padx=10, pady=5)
                continue
            if p.type == 'label':
                ttk.Label(self.top, text=p.label).pack(side=tk.TOP,fill=tk.BOTH,
                                                       expand=tk.TRUE, pady=10)
                continue

            v = getattr(cfg, p.name)()  # get value of current config parameter
            match p.type:
                case 'int':
                    self.vars[p.name] = tk.IntVar(value=v)
                case 'float':
                    self.vars[p.name] = tk.DoubleVar(value=v)
                case 'boolean':
                    self.vars[p.name] = tk.BooleanVar(value=v)
                case 'str':
                    self.vars[p.name] = tk.StringVar(value=v)

            if p.special == 'filechooser':
                self.create_dir_chooser_row(p.label, self.vars[p.name])
            else:
                self.create_config_row(p.label, self.vars[p.name], p.unit)

        btn_frame = ttk.Frame(self.top)
        ttk.Button(btn_frame, text="Close", command=self.close_dialog).pack(side=tk.RIGHT)
        ttk.Button(btn_frame, text="Apply", command=self.apply).pack(side=tk.RIGHT)
        
        config_frame.pack(side=tk.TOP, fill=tk.BOTH, expand=tk.TRUE)
        btn_frame.pack(side=tk.TOP, fill=tk.BOTH)

    def create_dir_chooser_row(self, label, variable):
        """Create a directory chooser config row."""

        container = ttk.Frame(self.top)
        container.pack(fill=tk.X, expand=tk.YES, pady=5)

        lbl = ttk.Label(master=container, text=label, width=35)
        lbl.pack(side=tk.LEFT, padx=5, fill=tk.X, expand=tk.NO)

        def _text_edit(event):
            """Keeps the config variable updated with text in the text widget."""
            event.widget.edit_modified(False)
            variable.set(event.widget.get('1.0', tk.END).rstrip())

        ent = tk.Text(container, wrap=tk.CHAR, width=35, height=4)
        ent.insert('1.0', variable.get())
        ent.pack(side=tk.LEFT, padx=5, fill=tk.X, expand=tk.YES)
        ent.bind('<<Modified>>', _text_edit)

        def _dir_chooser(variable):
            """Uses a filedialog to get a directory."""
            d = filedialog.askdirectory(parent=container, title='Select data directory',
                                        initialdir=variable.get())
            if d:
                ent.delete('1.0', tk.END)
                ent.insert('1.0', d)
                variable.set(d)

        btn = ttk.Button(container, text='...', width=3,
                         command=lambda: _dir_chooser(variable))
        btn.pack(side=tk.LEFT)

    def create_config_row(self, label, variable, unit):
        """Create a single row in the config dialog"""
        container = ttk.Frame(self.top)
        container.pack(fill=tk.X, expand=tk.YES, pady=5)

        lbl = ttk.Label(master=container, text=label, width=35)
        lbl.pack(side=tk.LEFT, padx=5, fill=tk.X, expand=tk.NO)

        if isinstance(variable.get(), bool):
            ent = ttk.Checkbutton(container, variable=variable)
        else:
            ent = ttk.Entry(master=container, textvariable=variable,
                            justify='right', width=15, font=('Arial 12'))
            unit = ttk.Label(master=container, text=unit, width=10)
            unit.pack(side=tk.RIGHT, padx=5)
        ent.pack(side=tk.RIGHT, padx=5, fill=tk.X, expand=tk.YES)

    def apply(self):
        for p in self.params:
            if p.name:
                getattr(cfg, p.name)(self.vars[p.name].get())

        if self.updated_cb:
            self.updated_cb()  # tell others that we've updated
        cfg.save_config()
        logger.info('Saved configuration')

    def close_dialog(self):
        self.top.destroy()